// VinciaMatching.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2015 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the Matching class. 
#include "VinciaMatching.h"

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  // init() : 
  // Initialize Matching interface
  //***********************************************************************

  bool Matching::init(int nVariationsIn) {

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("VM::init","Cannot initialize, pointers not set.");
      return false;
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Set number of uncertainty variations (at least one, for USR)
    nVariations = max(1,nVariationsIn);

    // Initialize Settings
    verbose          = settingsPtr->mode("Vincia:verbose");
    uncertaintyBands = settingsPtr->flag("Vincia:uncertaintyBands");
    
    // Initialize Shower Settings (needed to compute matching coefficients)
    evolutionType   = settingsPtr->mode("Vincia:evolutionType");
    orderingMode    = settingsPtr->mode("Vincia:orderingMode");
    sectorShower    = settingsPtr->flag("Vincia:sectorShower");
    alphaScmw       = settingsPtr->flag("Vincia:alphaScmw");    
    alphaSkMu       = settingsPtr->parm("Vincia:alphaSkMu");
    alphaSmode      = settingsPtr->mode("Vincia:alphaSmode");
    alphaSmodeNLO   = settingsPtr->mode("Vincia:alphaSmodeNLO");
    alphaSmuFreeze  = settingsPtr->parm("Vincia:alphaSmuFreeze");
    alphaSmax       = settingsPtr->parm("Vincia:alphaSmax");
    pTnorm          = settingsPtr->parm("Vincia:pTnormalization");
    mDnorm          = settingsPtr->parm("Vincia:mDnormalization");

    // Set matching alphaS: same as shower alphaS
    alphaSptr = &vinComPtr->alphaStrong;

    // Numerical settings
    nPointsMCmin    = settingsPtr->mode("Vincia:matchingNLOnPointsMin");
    nPointsMCmax    = settingsPtr->mode("Vincia:matchingNLOnPointsMax");
    nPointsMCmax    = max(nPointsMCmin,nPointsMCmax);

    // Default settings (for uncertainty variations reference setting)
    alphaScmwDef       = settingsPtr->flagDefault("Vincia:alphaScmw");    
    alphaSkMuDef       = settingsPtr->parmDefault("Vincia:alphaSkMu");
    alphaSmodeDef      = settingsPtr->modeDefault("Vincia:alphaSmode");
    alphaSmuFreezeDef  = settingsPtr->parmDefault("Vincia:alphaSmuFreeze");
    pTnormDef          = settingsPtr->parmDefault("Vincia:pTnormalization");
    mDnormDef          = settingsPtr->parmDefault("Vincia:mDnormalization");
    evolutionTypeDef   = settingsPtr->modeDefault("Vincia:evolutionType");
    orderingModeDef    = settingsPtr->modeDefault("Vincia:orderingMode");

    // Initialize internal diagnostic histograms
    // Histos for deltaA terms
    // TODO: move to common area and add histos for (1+V)
    // -> Vincia.info ?
    int    nBin   = 99;
    double xMin   = -0.099;
    double xMax   =  0.099;
    Hist Histo1("DeltaA/(1+V1)",nBin,xMin,xMax);
    for (int nGtmp = 0; nGtmp <= 9; ++nGtmp) {
      for (int nQQtmp = 0; nQQtmp <= 5; ++nQQtmp) {
        string state = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
        string hisTitle="DeltaA:" + state;
        HrDeltaA[hisTitle]=Histo1;
      }
    }

    // Check and initialize MG interface 
    isOnMG = settingsPtr->flag("Vincia:MGInterface");
    if (isOnMG) isOnMG = mgInterfacePtr->init();
    
    // Check and initialize BH interface
    isOnBH = false;

    // Return ok
    return isInit;
  }
	
  //***********************************************************************
  // Main functions for VINCIA matching
  //***********************************************************************

  // *********************************************************************
  // isMatched() : return true if matching exists at this order
  // *********************************************************************

  bool Matching::isMatched(int nIn, vector<int> idPartons, bool isPolarized,
			   bool treatAsMassless) {
    
    // Sanity check
    int nPartons = idPartons.size();
    if (nIn != 1) return false;
    if (nPartons <= 2) return false;

    // Return value false by default
    bool gotit = false;    

    // Data about resonance
    int colType  = particleDataPtr->colType(idPartons[0]);
    int chgType  = abs(particleDataPtr->chargeType(idPartons[0]));
    int spinType = particleDataPtr->spinType(idPartons[0]); 

    // Data about final state
    int nFS = idPartons.size() - nIn;
    int nG(0), nQ(0), nMisc(0);
    for (int i=nIn; i<int(idPartons.size()); ++i) {
      if (idPartons[i] == 21) ++nG;
      else if (idPartons[i] <= 8) ++nQ;
      else ++nMisc;
    }

    // Verbose output
    if (verbose >= 7) {
      cout<<" (VM::isMatched) state = ";
      for (int i=0; i<int(idPartons.size()); ++i) 
        cout<<idPartons[i]<<" ";
      cout<<endl;
      cout<<"      Resonance   : colT = "<<colType<<" chgT = "<<chgType
          <<" spinT = "<<spinType<<endl;
      cout<<"      Final State : treatAsMassless = "<<bool2str(treatAsMassless)
          <<" nQ = "<<nQ<<" nG = "<<nG<<" nU = "<<nMisc<<"   isPolarized = "
          <<bool2str(isPolarized)<<endl;      
    }
    
    // Helicity-dependence is switched off for massive particles
    if (!treatAsMassless) isPolarized = false;

    // So far matching only for pure gluon+quark final states
    if (nMisc >= 1) return false;
    
    // -------------------------------------------------------------------
    // Matching: with MG interface
    // -------------------------------------------------------------------
    if (isOnMG) {
      
      // Colorless resonances
      if (colType == 0) {

	// Scalar resonances (also used for undertermined spins)
	if (spinType <= 1) {	  

	  // Neutral (H0-like)
	  if (chgType == 0) {

            // H->qq, up to six partons
            if (nQ >= 2 && nFS <= 6) {
              gotit = true; 
              // Must be q-XXXX-qbar string.
              // Special case: check for Born-level H->gg; not matched
              if (particleDataPtr->colType(idPartons[nIn]) != 1) gotit = false;
              if (particleDataPtr->colType(idPartons[nPartons-1]) != -1) 
                gotit = false;              
            }

            // Pure gluons, up to H -> 5 gluons
            else if (nQ == 0 && nG <= 5) gotit = true;
            
          }

          // Charged (H+-like): only for helicity-dependent matching
	  else if (chgType == 3 && isPolarized) {

            // H->qq, up to six partons
            if (nQ >= 2 && nFS <= 6) gotit = true;
          }
          
        }
        
	// Vector resonances 
	else if (spinType == 3) {

	  // Neutral (Z-like)
	  if (chgType == 0) {
            
            // Z->qq, up to 6 jets
            if (nQ >= 2 && nFS <= 6) {
              gotit = true;
              // Must be q-XXXX-qbar string. Check for X -> gg
              if (particleDataPtr->colType(idPartons[nIn]) != 1) gotit = false;
              if (particleDataPtr->colType(idPartons[nPartons-1]) != -1) 
                gotit = false;              
            }
	  
          }

	  // Charged (W-like)
	  else if (chgType == 3) {
            
            // W->qq, polarized: up to 6 jets (using helamps from Z decay)
            if (isPolarized) {
              if (nQ >= 2 && nFS <= 6) gotit = true;
            }
            
            // W->qq, unpolarized: up to +2 gluons jets (using massive W)
            else {
              if (nQ == 2 && nG <= 2) gotit = true;              
            }
            
          }
          
        }
        
      }
      
    }

    // -------------------------------------------------------------------
    // Matching: with BlackHat interface
    // -------------------------------------------------------------------
    if (!gotit && isOnBH) {

      // So far no BlackHat matrix elements implemented
      gotit = false;

    }

    return gotit;

  } 

  // *********************************************************************
  // isMatchedNLO() : return true if NLO matching exists at this order
  // *********************************************************************
  
  bool Matching::isMatchedNLO(int nIn, vector<int> idParticles) {
    
    // Default is we don't have it
    bool gotIt = false;
    
    // Need at least 3 particles to be involved
    if (idParticles.size() <= 2) return false;
    
    // NLO matching only for decays so far
    if (nIn >= 2) return false;
    
    // Z decays
    if (idParticles[0] == 23) {
      
      // Z: NLO matching exists for Z->2,3 partons
      if (idParticles.size() == 3) gotIt = true;
      else if (idParticles.size() == 4) gotIt = true;
      
    }

    return gotIt;
    
  }

  // *********************************************************************
  // polarize() : use helicity matrix elements to polarize a list of 
  // partons 
  // *********************************************************************

  bool Matching::polarize(int nIn, vector<Particle>& partons) {

    // So far, we only know how to polarize resonance decays
    if (nIn != 1 || partons.size() < 3) return false;

    // Resonance decays
    if (nIn == 1) {

      // Identify unpolarized partonic state and check color ordering
      vector<int> idPartons;
      vector<Vec4> pPartons;
      vector<double> polPartons;
      bool isMassive = false;
      for (int i=0; i<int(partons.size()); ++i) {
	idPartons.push_back(partons[i].id());
	pPartons.push_back(partons[i].p());
	polPartons.push_back(9.);
	int acol = partons[i].acol();
	if (i >= 1 && acol > 0) { 
	  if (partons[i-1].col() != acol) {
	    if (verbose >= 3) 
	      printOut("VM::polarize",
		       "failed to polarize state. Not color ordered.");
	    return false;
	  }
        }
        // State should already have been mapped to massless at this
        // point (if massless) so only check for numerical rounding
        // errors to see if massive here or not.
        if (i >= nIn && pow2(pPartons[i].mCalc()/pPartons[i].e()) > TINY) 
          isMassive = true; 
      }
      if (verbose >= 6) 
        printOut("VM::polarize","isMassive = "+bool2str(isMassive));
      // Always allow decaying resonance to be polarized
      polPartons[0] = partons[0].pol();
      
      // Check if we have the relevant matrix element
      if (!isMatched(1,idPartons,true,!isMassive)) {
	if (verbose >= 6) 
	  printOut("VM::polarize",
		   "failed to polarize state. Unknown matrix element.");
	return false;      
      }
      
      // Get the polarization-summed ME
      double sumAll = mgInterfacePtr->ME2(1,idPartons,pPartons,polPartons);

      // Generate random number
      double sumR  = rndmPtr->flat() * sumAll;

      // Now loop over final-state polarizations until sum > sumR
      int npols = (int)pow(2.,(int)idPartons.size() - nIn);
      double sumNow = 0;
      for (int i = 0; i < npols; i++) {
	// Remove all but decaying resonance from polPartons,
	// then set this polarization state
	polPartons.resize(1);
	for (int j = 1; j <= (int)idPartons.size() - nIn; ++j) {	  
	  if (i % (int)pow(2.,j) < (int)pow(2.,j-1)) polPartons.push_back(-1.);
	  else polPartons.push_back(1.);
	}
	sumNow += mgInterfacePtr->ME2(1,idPartons,pPartons,polPartons);
	// Verbose output
	if (verbose >= 6) {
	  cout<<"(VM::polarize) Tested "<<num2str(polPartons[0],6)<<" -> ";
	  for (int j=1; j<= int(idPartons.size()) - 1; ++j) {
	    cout<<num2str(polPartons[j],6)<<" ";
	  }
	  cout<<" sumNow/sumAll = "<<sumNow/sumAll<<endl;	
	}
	// Check if this is the selected polarization
	if (sumNow >= sumR) break;
      }
      
      if (verbose >= 5) {
	cout<<"(VM::polarize) Chose FS polarizations = ";
	for (int j=1; j<= int(idPartons.size()) - 1; ++j) {
	  cout<<polPartons[j]<<" ";
	}
	cout<<endl;
      }

      // Save chosen polarizations to incoming parton list
      for (int i=nIn; i<int(partons.size()); ++i) {
	partons[i].pol(polPartons[i]);
      }

      // All is well
      return true;

    }

    // Should only arrive here if nIn != 1: failed.
    return false;
    
  }

  // Wrapper to operate directly on particles in an event record
  bool Matching::polarize(Event& event, int nIn, vector<int> iPartons) {
    
    // Sanity check
    if (nIn <= 0) return false;

    vector<Particle> partons;
    for (int i=0; i<int(iPartons.size()); ++i) {
      int iNow = iPartons[i];
      if (iNow <= 0 || iNow >= event.size()) return false;
      else partons.push_back(event[iNow]);
    }
    bool didPolarize = polarize(nIn,partons);
    if (didPolarize) {
      // Copy new polarizations
      for (int i=0; i<int(iPartons.size()); ++i) {
        int iNow = iPartons[i];
        event[iNow].pol(partons[i].pol()); 
      }
    }

    return didPolarize;
  }

  // *********************************************************************
  // getME() : Return a tree-level matrix element (squared)
  // INPUT: 
  //      nIn : 1 (for decay) or 2 (for 2->2 process)
  //      id  : list of particle PDG codes
  //            incoming first, then outgoing ones
  //      p   : same for momenta
  //      pol : same for polarizations
  //   Note: the particle list is assumed to be color ordered
  //
  // RETURN VALUE:
  //      -1 if no matrix element found, otherwise |M|^2
  // *********************************************************************

  double Matching::getME(int nIn, vector<int> id, vector<Vec4> p, 
			 vector<double> pol) {

    // Initialize to -1 return value
    double ME2 = -1.0;

    // First try MadGraph
    if (isOnMG) ME2 = mgInterfacePtr->ME2(nIn, id, p, pol);

    // Then try BlackHat 
    if (ME2 <= 0.0 && isOnBH) {
      // This part so far purely dummy. 
    }

    return ME2;

  }

  // Wrapper to use list of particles as input
  double Matching::getME(int nIn, vector<Particle> particles) {

    vector<int> id;
    vector<Vec4> p;
    vector<double> pol;
    for (int i=0; i<int(particles.size()); ++i) {
      id.push_back(particles[i].id());
      p.push_back(particles[i].p());
      pol.push_back(particles[i].pol());
    }
    
    return getME(nIn, id, p, pol);

  }

  // *********************************************************************
  // getPME() : Return a matching coefficient
  // INPUT: 
  //      nIn : 1 (for decay) or 2 (for 2->2 process)
  //      particles : vector containing particles
  //           first incoming particle(s), then outgoing ones
  //   Note   : the particle list is assumed to be color-ordered
  // OUTPUT: 
  //      PME : vector whose [0] component is the matching coefficient
  //            for the current shower settings, with the additional ones
  //            corresponding to uncertainty variations (if switched on).
  //              = |M_n|^2 / (Sum_i A^phys_i |M_n-1|^2)
  //      PME[0]   : reweighting factor with USR antennae
  //      PME[1]   : reweighting factor with DEF antennae 
  //      PME[2,3] : no difference at tree level (muR-variation)
  //      PME[4]   : reweighting factor with MAX antennae
  //      PME[5]   : reweighting factor with MIN antennae
  //      PME[6,7] : no difference at tree level (NLO K-factor variation)
  //      PME[8]   : reweighting factor: ordering variation 1 (split:pT)
  //      PME[9]   : reweighting factor: ordering variation 2 (emit:mD)
  //      PME[10]  : reweighting factor with all emission color factors = CA 
  //      PME[11]  : reweighting factor with all emission color factors = CF
  // *********************************************************************

  bool Matching::getPME(int nIn, vector<Particle> particles, 
			vector<double>& PME) {

    // Resize PME.
    PME.resize(0);
    bool done = false;

    // Matching using MadGraph
    if (! isOnMG) return false;

    // Create vectors of PDG ID codes, momenta, and polarizations
    vector<int> id;
    vector<Vec4> p;
    vector<double> pol; 
    vector<double> m;
    
    // Assign an ID code to the decaying resonance
    // Helicity-dependent: use Z for all vectors, H for all scalars
    // Unpolarized: use the actual ID code
    int id0 = 0;
    bool isMassless = true;
    bool isPolarized = true;
    for (int i=1; i<int(particles.size()); ++i) {
      if (particles[i].pol() == 9) isPolarized = false;
      if (particles[i].m() >= TINY) isMassless = false;
    }

    if (!isMassless || !isPolarized) {
      id0 = particles[0].id();
    }
    // Color singlets
    else if (particleDataPtr->colType(particles[0].id()) == 0) {
      // Spin-1 (e.g., W, Z, Z', ...)
      if (particleDataPtr->spinType(particles[0].id()) == 3) id0 = 23;
      // Spin-0 (e.g., h, h+, ...)
      if (particleDataPtr->spinType(particles[0].id()) == 1) id0 = 25;
    }

    if (verbose >= 8) 
      printOut("VM:getPME","Assigning decaying particle ID code "+num2str(id0));
    
    // If not known, return.
    if (id0 == 0) return -1;
    id.push_back(id0);
    p.push_back(particles[0].p());
    pol.push_back(particles[0].pol());
    m.push_back(particles[0].m());
    
    for (int i=1; i<int(particles.size()); ++i) {
      id.push_back(particles[i].id());
      p.push_back(particles[i].p());
      pol.push_back(particles[i].pol());
      m.push_back(particles[i].m());
    }

    /*
    cout<<"ME : pols = ";
    for (unsigned int i=0; i<id.size(); ++i) cout<<" "<<pol[i];
    cout<<endl;
    */

    // Get color-ordered M2 (= LC + ...) for this n-parton state
    double Mn2     = getME(1, id, p, pol);

    // mgInterface returns -1 if not existing. 
    if (abs(Mn2+1) < TINY) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }

    // SECTOR showers: find appropriate sector    
    // Loop over possible sectors, start with first one
    // Sector decomposition variables (see [LVS11]):
    //     10: QS2 = ARIADNE pT2 = sIK * yij * yjk 
    //      8: QS2 = sIK * yjk * sqrt(yij) (for jk collinear g -> qbar q)
    //      9: QS2 = sIK * yij * sqrt(yjk) (for ij collinear g -> qbar q)
    int irMin    = 2;
    if (sectorShower) {
      // Start with first clustering: ir = 2:
      double q2Min(0.);
      if (id[irMin] == 21) 
	q2Min = resolutionPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],10);
      else if (id[irMin] < 0) 
	q2Min = resolutionPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],8);
      else 
	q2Min = resolutionPtr->Q2E(p[irMin-1],p[irMin],p[irMin+1],9);
      // Loop over additional clusterings, ir >= 3:
      for (int ir=3; ir<int(p.size())-1; ++ir) {
	double q2(0.);
	if (id[ir] == 21) 
	  q2 = resolutionPtr->Q2E(p[ir-1],p[ir],p[ir+1],10);
	else if (id[ir] < 0)
	  q2 = resolutionPtr->Q2E(p[ir-1],p[ir],p[ir+1],8);
	else 
	  q2 = resolutionPtr->Q2E(p[ir-1],p[ir],p[ir+1],9);
	if (q2 < q2Min) {
	  q2Min = q2;
	  irMin = ir;
	}
      }    
      if (verbose >= 8) 
	cout<<"VM: chose sector "<<irMin<<" with clusterer id "
	    <<id[irMin]<<endl;
    }

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Parton to be clustered
      int ir    = 2+iClu;

      // SECTOR showers: skip all clusterings except correct sector
      if (sectorShower && ir != irMin) continue;

      // Verbose
      if (verbose >= 8) printOut("VM::getPME","Doing clustering no. "
                                 +num2str(iClu));
      
      // Compute basic quantities
      double ma = m[ir-1];
      double mr = m[ir];
      double mb = m[ir+1];
      double sP  = m2(p[ir-1]+p[ir]+p[ir+1]);
      double mP  = sqrt(sP);
      double mMua = ma / mP;
      double mMur = mr / mP; 
      double mMub = mb / mP;
      
      //Check to see if any particles are unpolarized
      //If so, then we match to fully unpolarized matrix elements
      //Ignore initial decaying particle
      bool isPol = true;
      for (int ipol = 1; ipol < (int)pol.size(); ipol++) {
        if (pol[ipol] > 8.0) {
          isPol = false;
          break;
        }
      }

      // Reconstruct previous state
      vector<int> idClu;
      vector<Vec4> pClu;
      vector< vector<int> > polClu;
      int antIndx=-1;
      bool swapAB=false;
      if (isPol) {
	// Determine all polarized clusterings as recorded in polClu
	// Note: if any of the particles are unpolarized -> unpolarized
	vinCluPtr->getCluPol(id,p,ir-1,ir,ir+1,idClu,pClu,antIndx,swapAB,
			     pol,polClu);
      } else {
	vinCluPtr->getClu(id,p,ir-1,ir,ir+1,idClu,pClu,antIndx,swapAB);
      }

      // Invariant mass of parent antenna and daughter invariants
      double yar = (m2(p[ir-1],p[ir])-pow2(ma)-pow2(mr))/sP;
      double yrb = (m2(p[ir],p[ir+1])-pow2(mr)-pow2(mb))/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
        // MR 17.1.2011: have to swap masses as well
        double mMutmp = mMua;
        mMua = mMub;
        mMub = mMutmp;
      }

      // Check for vanishing invariants
      if (yar < TINY || yrb < TINY) {
	printOut("VM::getPME","Warning: small invariant(s): y = "
		 +num2str(yar)+","+num2str(yrb)
		 +" sParent ="+num2str(sP));
	if (verbose >= 3) {
	  cout<<"Momentum Listing:"<<endl;
	  cout<<p[ir-1]<<p[ir]<<p[ir+1];
	  cout<<"Invariants: "<<m2(p[ir-1]+p[ir])<<" "<<m2(p[ir]+p[ir+1])<<endl;
	}
      }
      
      // Compute Pimp factor (only used for global showers) 
      //  : orderingMode <= 1 : Pimp = 1
      //  : orderingMode == 2 : Q2E/(Q2Ehat+Q2E) 
      double Pimp         = 1.0;
      // Additional Pimp factor variations used for uncertainty estimates
      double PimpDef      = 1.0;
      double PimpPT       = 1.0;
      double PimpMD       = 1.0;

      // Only apply if at least 3 partons in final state
      if ( !sectorShower && orderingMode >= 2 && int(pClu.size()-2) > nIn) { 

	// QE, pT, and mD Values for current clustering	
	double QE2new, QE2newDef, pT2new, mD2new;
	// If this is an anti fermion -> always use m2rb for g->qqbar
	if (id[ir] < 0) {
          double m2rb = m2(p[ir],p[ir+1]); 
          QE2new      = m2rb;
          QE2newDef   = m2rb;
          mD2new      = m2rb;
          pT2new      = resolutionPtr->yE(yar,yrb,1)*sP;
        }
	// Else if this is a fermion -> always use m2ar for g->qqbar
	if (id[ir] < 0) {
          double m2ar = m2(p[ir-1],p[ir]); 
          QE2new      = m2ar;
          QE2newDef   = m2ar;
          mD2new      = m2ar;
          pT2new      = resolutionPtr->yE(yar,yrb,1)*sP;
	}
	// Else use QE, pT, mD definitions for g emission
	else {
	  QE2new = resolutionPtr->yE(yar,yrb)*sP;
	  QE2newDef = pTnormDef * yar * yrb * sP;
          mD2new    = resolutionPtr->yE(yar,yrb,2)*sP;
          pT2new    = resolutionPtr->yE(yar,yrb,1)*sP;
	}

	// Initial values for Hat variables characterizing the reclustered
	// parton state
	double QE2hat    = m2(pClu[1],pClu[2],pClu[3]);
        double QE2hatDef = QE2hat;
        double pT2hat    = QE2hat;
        double mD2hat    = QE2hat;

	// Loop over possible clusterings, find smallest Q for each measure
	for (int jClu=0; jClu<int(pClu.size())-3; jClu++) {
	  int jaHat=jClu+1;
	  int jrHat=jClu+2;
	  int jbHat=jClu+3;
	  // If this is an anti fermion -> use m2rb for g->qq splitting
	  if (idClu[jrHat] < 0 ) {
            double m2rb = m2(pClu[jrHat],pClu[jbHat]);
	    QE2hat      = min(QE2hat,m2rb);
	    QE2hatDef   = min(QE2hatDef,m2rb);
	    mD2hat      = min(mD2hat,m2rb);
            pT2hat      = min(pT2hat,resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],
                                                        pClu[jbHat],1));
          }
	  // Else if this is a fermion, use m2ar for g->qq splitting
          else if (idClu[jrHat] < 20) {
            double m2ar = m2(pClu[jaHat],pClu[jrHat]);
            QE2hat      = min(QE2hat,m2ar);
            QE2hatDef   = min(QE2hatDef,m2ar);
	    mD2hat      = min(mD2hat,m2ar);
            pT2hat      = min(pT2hat,resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],
                                                        pClu[jbHat],1));
          }
	  // Else use individual QE, pT, or mD measure for g emission
	  else {
	    double QE2    = 
              resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat]);
            double pT2    = 
              resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],1);
	    double mD2    = 
              resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],2);
	    double QE2def = pTnormDef/pTnorm * pT2;
            QE2hat    = min(QE2hat,QE2);
	    QE2hatDef = min(QE2hatDef,QE2def);
            mD2hat    = min(mD2hat,mD2);
            pT2hat    = min(pT2hat,pT2);
	  }
	}	  
	// Standard Pimp factor
	Pimp         = QE2hat/(QE2hat+QE2new);
	PimpDef      = QE2hatDef/(QE2hatDef+QE2newDef);
	// Ordering variations : pT and mD
	PimpPT       = pT2hat/(pT2hat+pT2new);
	PimpMD       = mD2hat/(mD2hat+mD2new);
	if (verbose >= 8) 
	  printOut("VM::getPME","Pimp = "+num2str(Pimp));
      }
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (!sectorShower && (antIndx == iQGsplit || antIndx == iGGsplit)) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  if (ir+1 < int(pClu.size()) ) { 
	    sN = m2(pClu[ir],pClu[ir+1]);
	    // Only apply if neighbor is a quark
	    if (idClu[ir+1] != 21) 
	      Pari = 2.0*sN/(sP + sN);
	  }
	} else { 
	  // Only apply if neighbor is in final state
	  if (ir-2 >= nIn) {
	    sN = m2(pClu[ir-2],pClu[ir-1]);	
	    // Only apply if neighbor is a quark
	    if (idClu[ir-2] != 21)  
	      Pari = 2.0*sN/(sP + sN);
	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("VM::getPME","Pari = "+num2str(Pari));
	
      }
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      //Here, need to sum over possible spin clustering histories
      //only sum over multiple histories if all particles are polarized
      int polClust = 1;
      if (isPol) {
        polClust = (int)polClu.size();
        if (verbose >= 8) 
          cout<<" Looping over "<<polClust<<" internal helicities"<<endl;
      }
      for (int ipol = 0; ipol < polClust; ++ipol) {
        double Mp2;
	if (!isPol) Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu);
	else Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu, polClu[ipol]);
	if (Mp2 <= 0.0) {
	  string idString=num2str(idClu[0])+" -> ";
	  string polString=num2str(polClu[ipol][0])+" -> ";
	  for (unsigned int i=1;i<idClu.size();i++) 
	    idString+=num2str(idClu[i]);
	  for (unsigned int i=1;i<polClu[ipol].size();i++) 
	    polString+=num2str(polClu[ipol][i]);
	  printOut("VM::getPME","Unphysical momenta or unknown ME "
		   "for "+idString);
	  printOut("VM::getPME","                                 "
		   "hel "+polString);
	  // Print momentum configuration
	  if (verbose >= 3) {
	    for (unsigned int i=0;i<pClu.size();i++) {
	      cout<< "    i = "<<i<<" px = "<<num2str(pClu[i].px())<<" py = "
		  <<num2str(pClu[i].py())<<" pz = "<<num2str(pClu[i].pz())<<
		" e = "<<num2str(pClu[i].e())<<" m = "<<pClu[i].mCalc()<<endl;
	    }
	  }
        }
      
        // TMP!
        if (verbose >= 9) 
          printOut("VM::getPME","Mp2 = "+num2str(Mp2));
	
        // Check if return was -1 ?
        if (abs(Mp2+1.0) < TINY) return false; 
      
        // Loop over different antenna sets (for uncertainty evaluation)
        // Note: we do not make modifications for strong ordering here, since
        // we assume unordered generation up to the matched orders, and hence
        // the matching corrections should always include unordering. 
        // We also do not make variations in alphaS here, since all powers
        // of the coupling are factored out of the matrix element correction.

	// Set polarizations	
	int polA(9),polB(9),poli(9),polj(9),polk(9);
	if (isPol) {
	  polA = (!swapAB) ? polClu[ipol][ir-1] : polClu[ipol][ir];
	  polB = (!swapAB) ? polClu[ipol][ir] : polClu[ipol][ir-1];
	  poli = (!swapAB) ? pol[ir-1] : pol[ir+1] ;
	  polj = pol[ir];
	  polk = (!swapAB) ? pol[ir+1] : pol[ir-1];
	}

        // Compute antenna function values for USR antennae and optionally
        // for uncertainty variations. Use DEF as basis for variations.
        vector<double> antVal;
        antVal.push_back(getAnt(antIndx,sectorShower,0)->
                          antennaFunction(yar,yrb,mMua,mMur,mMub,
                                          polA,polB,poli,polj,polk)/sP); 
        if (uncertaintyBands) {
          antVal.push_back(getAnt(antIndx,sectorShower,1)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
          antVal.push_back(getAnt(antIndx,sectorShower,2)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
          antVal.push_back(getAnt(antIndx,sectorShower,3)->
                            antennaFunction(yar,yrb,mMua,mMur,mMub,
                                            polA,polB,poli,polj,polk)/sP); 
        }

        // Loop over uncertainty variations (0 = USR, 1 = DEF, ...)
        for (int iVar=0; iVar<nVariations; iVar++) {	

          // Get saved value of physical antenna function
          double antPhys = antVal[0];
          if (iVar >= 1) antPhys = antVal[1];
          if (iVar == 4) antPhys = antVal[2];
          if (iVar == 5) antPhys = antVal[3];

          // Color/charge factor (USR, DEF, and color-factor variations)
          double chargeFactor = 
            getAnt(antIndx,sectorShower,0)->getChargeFactor();
          if (iVar >= 1) chargeFactor = 
                           getAnt(antIndx,sectorShower,1)->getChargeFactor();
          if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
          if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	  
	  // Multiply by charge, Pimp, and Pari factors
          if (iVar == 0) 
            antPhys *= Pimp * Pari * chargeFactor;  
          else 
            antPhys *= PimpDef * Pari * chargeFactor;  

	  // Apply pimp factor modifications to variations 8 and 9
	  if (!sectorShower) {
	    if (iVar == 8) antPhys *= PimpPT/PimpDef;
	    if (iVar == 9) antPhys *= PimpMD/PimpDef;
	  }         
              
          // verbose output
          if ( (verbose >= 8 && iVar == 0) || verbose >= 9) {
            cout<<"VM: antPhys("<<polA<<polB<<">"<<poli<<polj<<polk<<") = "
                <<antPhys<<", isSwapped = "<<bool2str(swapAB)
                <<" m(ar,rb) = "<<sqrt(yar*sP)<<", "<<sqrt(yrb*sP);
            if (iVar >= 1) cout <<" (iVar = "<<iVar<<")"<<endl;
            cout<<endl;
          }

	  // Add to Approximated Mn2 for this set
	  Mn2Appr[iVar] += Mp2 * antPhys;
        }
        
        //End loop over spin clusterings
      }
      
    }
    
    // Store ME reweight factors
    for (int iVar=0; iVar<nVariations; iVar++) {
      if (iVar >= 1 && ! uncertaintyBands) continue;
      PME.push_back(Mn2/Mn2Appr[iVar]);
    }

    // Diagnostics if shower VERY bad approximation
    if (PME[0] > 1000.0 || (verbose >= 3 && PME[0] > 100.0 )) {
      printOut("VM::getPME","Warning: large correction at nLegs ="
	       +num2str(int(p.size()-1),2)
	       +" unregulated PME[0] = "+num2str(PME[0])
	       +" ME"+num2str(int(p.size()-1),1)+" = "+num2str(Mn2)
	       +" Sum(PS*ME"+num2str(int(p.size()-2),1)+") = "+num2str(Mn2Appr[0]));
    }

    // Store Old ME (future possibility for speedup)
    // oldME2 = Mn2;
   
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

  //*******

  // getSudVarEmit():

  // Get difference in Sudakov integral between two different evolution
  // types. Used for uncertainty estimates (eg estimate mD Sudakov when
  // using pT ordering). Estimate is based on transcendental parts of A31.

  /*
  double Matching::getSudVarEmit(int iVar, double y12, double y23, 
                                 double sAnt) {

    // This uncertainty only needs to be computed when changing to 
    // ev variable with different contours. 
    if (iVar <= 7 || iVar >= 10) return 0.0;

    double diff = 0.;
    int ordeMode = settingsPtr->("Vincia:orderingMode");
    int evType   = settingsPtr->("Vincia:evolutionType");    

    // Variation 8: return difference between mD and pT evolution 
    // (hardcoded for the case of qq -> qgqbar, assumed to be good
    // enough for uncertainty estimates)
    if (iVar == 8) {
      //VinciaSubtractionTerms Sud23St;
      double Sud23St, Sud34GESm, Sud34GSSm;        
      double Sud23StMD(0.), Sud34GESmMD(0.), Sud34GSSmMD(0.);
      double Sud23StPT(0.), Sud34GESmPT(0.), Sud34GSSmPT(0.);
      
      // 2->3 Emission Sudakov expansions
      // Integrate emission probability from antenna mass down
      // to pT or mD scale of emission. 
      Sud23emitPT   = subtractionTerms.IqqEmit(yij,yjk,1);
      Sud23emitMD   = subtractionTerms.IqqEmit(yij,yjk,2);
      diff = Sud23StPT - Sud23StMD;
      
      // 3->4 Emission Sudakov expansions
      Sud34emitPT = subtractionTerms.IqgEmit(yij,yjk,1,ordeMode);
      Sud34emitMD = subtractionTerms.IqgEmit(yij,yjk,2,ordeMode);
      diff += Sud34emitPT - Sud34emitMD;

      // 3->4 Splitting Sudakov expansions (with splittings ordered in mqq)
      Sud34GSSm = subtractionTerms.IqgSplit(yij,yjk,1,ordeMode,normFacPT);
        if (uncertaintyBands) {
          // Not calculating 34 splitting Sudakov for Q(g->qq) = pT yet -> Lisa
          Sud34GSSmPT = 0.0;
          Sud34GSSmMD = subtractionTerms.IqgSplit(yij,yjk,2,ordeMode,1.);
        }

        if( verbose >= 9 ){
          printOut("VM::getVME()","2->3 Shower integrals = "
                                                       +num2str(Sud23St));
          printOut("VM::getVME()","3->4 GE Shower integrals = "
                                                       +num2str(Sud34GESm));
          printOut("VM::getVME()","3->4 GS Shower integrals = "
                                                       +num2str(Sud34GSSm));
	}

        // Building the expanded shower sudakov
        double shower   = Sud23St - Sud34GESm - Sud34GSSm;
        double showerMD = Sud23StMD - Sud34GESmMD - Sud34GSSmMD;
        double showerPT = Sud23StPT - Sud34GESmPT - Sud34GSSmPT;
      
    }

    // Variation 9: return difference between using mqq or pT evolution
    // for subsequent splittings of the emitted gluon to quarks
    // (hardcoded for the case of qq -> qgqbar, assumed to be good
    // enough for uncertainty estimates)

  }

  */

  //*******

  // getVME():

  // Get a single VME coefficient, for:
  //    VME[0]        = matching factor for user settings
  //    VME[1]        = -"- for default settings
  //    VME[2]        = -"- for muRdef/2
  //    VME[3]        = -"- for muRdef*2
  //    VME[4]        = VME[1] (fixed by LO matching)
  //    VME[5]        = VME[1] (fixed by LO matching)
  //    VME[6]        = VME[1] * (1+NNLO)
  //    VME[7]        = VME[1] * (1-NNLO)
  //    VME[8]        = ordering Var 1 (split:pT)
  //    VME[9]        = ordering Var 2 (emit:mD)
  //   VME[10]        = VME * (1+1/NC^2)
  //   VME[11]        = VME / (1+1/NC^2)

  bool Matching::getVME(int nIn, vector<Particle> particles, 
                       vector<double>& VME) {

    // Default value = 1
    bool status = false;
    
    // Sanity checks
    if (nIn != 1) return status;

    // Require at least 3 particles
    if (particles.size() <= 2) return status;

    // Sanity check
    double eCM = particles[0].mCalc();
    if (eCM <= 0.0) {
      printOut("VM::getVME","ERROR! eCM <= 0");
      return false;
    }

    // We still only know how to compute this for massless particles
    for (int i = 1; i < int(particles.size()); ++i) 
      if (pow2(particles[i].m()/eCM) > TINY) {
        if (verbose >= 9) {
          cout<<"  NLO : i = "<<i<<" m = "<<particles[i].m()<<" m2/s = "
              <<pow2(particles[i].m()/eCM)<<endl;
          cout<<"   p "<<particles[i].p();
        }
        return false;
      }
    
    // Settings extracted from Vincia
    int evTypeEmit = evolutionType;
    int ordeMode   = orderingMode;
    double evFacN  = 1.0;
    if (evTypeEmit == 1) evFacN = pTnorm;
    else if (evTypeEmit == 2) evFacN = mDnorm;

    // Data about final state
    int idRes     = particles[0].id();
    int nFin      = particles.size() - nIn;

    // Global Invariants
    double mSys   = particles[0].mCalc();
    double m2sys  = pow2(mSys);

    // Determine mu2me: renormalization scale for NLO correction term
    // Nominally chosen to be mSys (~ hard correction) 
    // (or geometric mean of s(ijk) for nFin >= 4).
    // Alternatively compute muEff as geometric mean of all pT values
    double mu2me  = m2sys;
    int nFac      = 0;
    if (nFin >= 3) {
      for (int iClu=0; iClu<nFin; ++iClu) {
        int ia = nIn + iClu - 1;
        int ir = nIn + iClu;
        int ib = nIn + iClu + 1;
        // Only attempt clustering of first particle if gluon (cyclic)
        if (iClu == 0) {
          if (particles[1].colType() != 2) continue;
          else ia = nIn+nFin-1;
        }
        // Only attempt clustering of last particle if gluon (cyclic)
        else if (iClu == nFin-1) {
          if (particles[nIn+nFin-1].colType() != 2) continue;
          else ib = nIn;
        }      
        double sIK = m2(particles[ia].p() + particles[ir].p()
                        + particles[ib].p());        
        double mu2clu = sIK;
        if (alphaSmodeNLO == 1) { 
          double sij = 2*particles[ia].p()*particles[ir].p();
          double sjk = 2*particles[ir].p()*particles[ib].p();
          mu2clu = sij*sjk/sIK;        
        } 
        if (nFac == 0) mu2me = mu2clu;
        else mu2me *= mu2clu;
        nFac += 1;
      }
      mu2me = pow(mu2me,1./nFac); 
    }
    // Limit mu2me scale from below
    mu2me = mu2me + pow2(alphaSmuFreezeDef);    
    // Also set nFme: number of flavors at ME scale
    double nFme      = vinComPtr->getNf(sqrt(mu2me));
    
    // AlphaS factor to use for corrections (at ME scale)
    double alphaS = min(alphaSmax,alphaSptr->alphaS(mu2me));
    double alphaSfac = alphaS/2./M_PI;

    // Z decays
    if (idRes == 23) {

      // Z -> 2
      if (nFin == 2){ 
        VME[0]=alphaS/M_PI;
        status = true;
      }

      // Z -> 3
      else if (nFin == 3) {

        // Invariants : Q_I G_J Qbar_K
        double sij = m2(particles[1].p() + particles[2].p());
        double sjk = m2(particles[2].p() + particles[3].p());
        double sIK = m2sys;
        double yij = sij/m2sys;
        double yjk = sjk/m2sys;
        
	// NLO matching

	//Fixed Order contribution
	double Born,RfunFull1,RfunFull2,RfunFull3;

	Born = yij/yjk + (2*(1 - yij - yjk))/(yij*yjk) + yjk/yij;
	RfunFull1 = (nFme*(log(yij) + log(yjk)))/6.;
	RfunFull2 = CA*(-4 - pow2(M_PI)/6. - (5*log(yij))/3. + 
                        log(1 - yij)*log(yij) - (5*log(yjk))/3. 
                        - log(yij)*log(yjk) + log(1 - yjk)*log(yjk) + 
			Li2(yij) + Li2(yjk));
	RfunFull3 =(CA*(1 + (1 - yij)/(2.*yij) - yij/(2.*(1 - yjk)) 
                        + (1 - yjk)/(2.*yjk) - yjk/(2.*(1 - yij)) 
                        + (2 + (2*(yij - yjk))/(1 - yij) - 
                           (yij*yjk)/(2.*pow2(1 - yij)))*log(yij) + 
                        (2 - (yij*yjk)/(2.*pow2(1 - yjk)) + 
                         (2*(-yij + yjk))/(1 - yjk))*log(yjk)))/Born;

        if( verbose == 9 ){
          printOut("VM::getVME","Loop calculation = "
                   +num2str(RfunFull1+RfunFull2+RfunFull3));
	}
        // Sum up Rfun pieces from one-loop matrix element
        double RfunFull = RfunFull1 + RfunFull2 + RfunFull3;

	// Renormalization scale of parton shower
        double mu2ps = sIK;
        if (alphaSmode == 1) mu2ps = sij*sjk/sIK;
        else if (alphaSmode == 2) mu2ps = min(sij,sjk);
        mu2ps *= pow2(alphaSkMu);
        // Default renormalization scale (for uncertainty variation reference)
        // and factor 2 variations
        double mu2def = sIK;
        if (alphaSmodeDef == 1) mu2def = sij*sjk/sIK;
        else if (alphaSmodeDef == 2) mu2def = min(sij,sjk);        
        mu2def *= pow2(alphaSkMuDef);
        double mu2min = mu2def / 4.;
        double mu2max = mu2def * 4;
	// Number of flavors used in parton-shower alphaS
	double nFps = vinComPtr->getNf(sqrt(mu2ps));

        // Translate from CMW to MSbar scheme
        // (if shower uses CMW and assuming ME uses MSbar)
        double facCMW2 = pow2(vinComPtr->alphaStrong.facCMW(nFps));
        if (alphaScmw) mu2ps /= facCMW2;
        if (alphaScmwDef) {
          mu2def /= facCMW2;
          mu2max /= facCMW2;
          mu2min /= facCMW2;
        }

        // Renormalization logs (including scale variations)
        double SLmuR    = (11/2. - nFme/3.)*log(sIK/mu2ps)*alphaSfac;
        double SLmuRdef = (11/2. - nFme/3.)*log(sIK/mu2def)*alphaSfac;
        double SLmuRmin = (11/2. - nFme/3.)*log(sIK/mu2min)*alphaSfac;
        double SLmuRmax = (11/2. - nFme/3.)*log(sIK/mu2max)*alphaSfac;

	// PS: 08 Jul 2013
	// Add threshold logs if numbers of ME and PS flavors not equal
	// (formally arise from translating PS alphaS to ME one)
        // Check that sign change for thresholds above is ok!
        // Text currently claims it is ok.
	if (nFps < nFme) { 
	  for (int idQ = nFme; idQ > nFps; --idQ) {
	    double m2thres = alphaSptr->muThres(idQ);
	    SLmuR    += 1/3. * log(m2thres/min(mu2ps,m2thres)) * alphaSfac;
	    SLmuRdef += 1/3. * log(m2thres/min(mu2def,m2thres)) * alphaSfac;
	    SLmuRmin += 1/3. * log(m2thres/min(mu2min,m2thres)) * alphaSfac;
	    SLmuRmax += 1/3. * log(m2thres/min(mu2max,m2thres)) * alphaSfac;
	  }
	}
	else if (nFps > nFme) {
	  for (int idQ = nFme+1; idQ <= nFps; ++idQ) {
	    double m2thres = alphaSptr->muThres(idQ);
	    SLmuR    -= 1/3. * log(m2thres/max(mu2ps,m2thres)) * alphaSfac;
	    SLmuRdef -= 1/3. * log(m2thres/max(mu2def,m2thres)) * alphaSfac;
	    SLmuRmin -= 1/3. * log(m2thres/max(mu2min,m2thres)) * alphaSfac;
	    SLmuRmax -= 1/3. * log(m2thres/max(mu2max,m2thres)) * alphaSfac;
	  }
	}
	
	//VinciaSubtractionTerms Sud23St;
	double Sud23St, Sud34GESm, Sud34GSSm;        

        // 2->3 Emission Sudakov expansion
        Sud23St   = subtractionTermsPtr->IqqEmit(yij,yjk,evTypeEmit);

        // 3->4 Emission Sudakov expansion
        Sud34GESm = subtractionTermsPtr->IqgEmit(yij,yjk,evTypeEmit,ordeMode);

        // 3->4 Splitting Sudakov expansion 
        Sud34GSSm = subtractionTermsPtr->IqgSplit(yij,yjk,evTypeEmit,ordeMode,
                                              evFacN);
        // Verbose output
        if( verbose >= 9 ){
          printOut("VM::getVME()","2->3 Shower integrals = "
                                                       +num2str(Sud23St));
          printOut("VM::getVME()","3->4 GE Shower integrals = "
                                                       +num2str(Sud34GESm));
          printOut("VM::getVME()","3->4 GS Shower integrals = "
                                                       +num2str(Sud34GSSm));
	}

        // Building the expanded shower sudakov
        double shower   = Sud23St - Sud34GESm - Sud34GSSm;

        // born level matching
        double V0 = 2.;
        shower   -= V0;
        
        // finite terms arising from integrating antenna over full phase space
	// (using same number of flavors, nF, as ME correction)
        double intd30, inte30;
        intd30 = 34/3.*CA;
        inte30 = -nFme;        
        shower   += intd30 + inte30;

	// Ariadne log
	double AriadneLog;
	AriadneLog = -nFme*(yij-yjk)/(yij+yjk)*log(yij/yjk)/6.;
        shower   += AriadneLog;

        if( verbose == 9 ){
          printOut("VM::getVME()","AriadneLog = "+num2str(AriadneLog));
	}

        // Finite terms arising from the tree-level matching corrections
        // (the deltaA terms, representing the deviation from standard GGG)
        // TODO: should be delegated to a separate function
        double sumEmit     = 0.;
        double sumSplit    = 0.;
        double comFacEmit  = CA/4.;
        double comFacSplit = nFme/4.;

        // Store original momentum configuration as 4-vectors
        vector<Vec4> pOrg;
        for (int iOrg=1; iOrg<=3; ++iOrg) 
          pOrg.push_back(particles[iOrg].p());
        double antOrg = functionsPtr->A30(pOrg,0,1,2);

        // Verbose output
        if (verbose >= 9) {
          cout<<"\n\n pOrg:\n"<<pOrg[0]<<pOrg[1]<<pOrg[2]<<endl;
          cout<<" antOrg = "<<antOrg<<endl;
          cout<<"=== Beginning MC Loop ==== "<<endl;
        }

        // Do MC integration loop
        double precision = 1.0;
        int    nPointsMC = 0;
        double deltaA    = 0.0;
        double deltaAsq  = 0.0;
        double twoPi     = 2. * M_PI;
        double phi       = 0.;
        vector<double> x;
        // Make VinClu less verbose to avoid cluttering output
        vinCluPtr->setVerbose(verbose - 2);
	double precisionGoal = 0.01;
        while (precision > precisionGoal) {
          ++nPointsMC;

          // The n-parton phase-space point already uses yij and yjk
          // Use xij and xjk to integrate over (n+1)-parton phase space
          x.resize(0);
          x.push_back(rndmPtr->flat());
          x.push_back(rndmPtr->flat());
          if (x[0] + x[1] > 1.0) {
            x[0] = 1. - x[0];
            x[1] = 1. - x[1];
          }
          phi = twoPi * rndmPtr->flat();

          // Verbose output
          if (verbose >= 9) {
            cout<<" Point "<<nPointsMC<<" scaled invariants:"<<endl;
            cout<<" x[0] = "<<x[0]<<" x[1] = "<<x[1]<<endl;
            cout<<"--- Beginning sum over radiating antennnae ---"<<endl;
          }

          // Sum over radiating antennae.
          vector<Vec4> nPlusOneE, nPlusOneS;
          double thisEmit  = 0.;
          double thisSplit = 0.;
          for (int iA = 0; iA < int(pOrg.size()) -1 ; ++iA) {

            // Set up radiating system: [iA,iA+1]
            vector<Vec4> pTwo;
            pTwo.push_back(pOrg[iA]);
            pTwo.push_back(pOrg[iA+1]);
            double sAnt = m2(pTwo[0],pTwo[1]);
            
            //-------------------------------
            // QG Emit (and GQbar, with swap)
            int iAntE = iQGemit;
            int iAntS = iQGsplit;
            bool swap = false;
            if (iA == 1) swap = true;

            // Compute antennae. Conjugate if needed.
            Antenna* antEptr = antennaSets[0]->getAnt(iAntE);
            Antenna* antSptr = antennaSets[0]->getAnt(iAntS);
            double aVinciaE  = antEptr->antennaFunction(x[swap],x[1-swap])/sAnt;
            double aStdE     = subtractionTermsPtr->getA(iAntE,x[swap],
                                                     x[1-swap])/sAnt;
            double aVinciaS  = antSptr->antennaFunction(x[swap],x[1-swap])/sAnt;
            double aStdS     = subtractionTermsPtr->getA(iAntS,x[swap],
                                                     x[1-swap])/sAnt;

            // Evaluate Pimp factors
            double ordE   = 1.0;
            double ordS   = 1.0;
            // qg parent
            if (iA == 0) {  
              if (evolutionType == 1) {
                ordE = yjk/(yjk + x[0]*x[1]);
                ordS = yjk/(yjk + x[1]/pTnorm);
              } else {
                ordE = min(yij,yjk)/(min(yij,yjk) + min(x[0],x[1])*yij);
                ordS = min(yij,yjk)/(min(yij,yjk) + x[1]*yij/mDnorm); 
              }
            }
            // gqbar parent
            else {
              if (evolutionType == 1) {
                ordE = yij/(yij + x[0]*x[1]);
                ordS = yij/(yij + x[0]/pTnorm);
              } else {
                ordE = min(yij,yjk)/(min(yij,yjk) + min(x[0],x[1])*yjk);
                ordS = min(yij,yjk)/(min(yij,yjk) + x[0]*yjk/mDnorm);
              }
            }

            //Verbose output
            if (verbose >= 9) {
              cout<<" Antenna "<<iA<<endl;
              cout<<pTwo[0]<<pTwo[1];
              cout<<" aVinciaE = "<<aVinciaE<<" ordE = "<<ordE
                  <<" aStdE = "<<aStdE<<endl;
              cout<<" aVinciaS = "<<aVinciaS<<" ordS = "<<ordS
                  <<" aStdS = "<<aStdS<<endl;
            }

            // Construct (n+1)-parton phase-space point
            int kineMapE = antEptr->getKineMapType();
            int kineMapS = antSptr->getKineMapType();
            vector<Vec4> pThreeE, pThreeS;            
            vinCluPtr->map2to3light(pThreeE,pTwo,kineMapE,x[0],x[1],phi);
            if (kineMapS == kineMapE) {
              pThreeS = pThreeE;
            } else {
              vinCluPtr->map2to3light(pThreeS,pTwo,kineMapS,x[0],x[1],phi);
            }
            // Store new particles in vector that can be passed the LO ME
            nPlusOneE.resize(0);
            for (int iCopy=0; iCopy<int(pOrg.size()); ++iCopy) {
              if (iCopy != iA && iCopy != iA+1) {
                nPlusOneE.push_back(pOrg[iCopy]);
                nPlusOneS.push_back(pOrg[iCopy]);
              }
              else if (iCopy == iA) {
                // Insert branched momenta in place of unbranched ones
                nPlusOneE.push_back(pThreeE[0]);
                nPlusOneE.push_back(pThreeE[1]);
                nPlusOneE.push_back(pThreeE[2]);
                nPlusOneS.push_back(pThreeS[0]);
                nPlusOneS.push_back(pThreeS[1]);
                nPlusOneS.push_back(pThreeS[2]);                
              }
            } 
            // Numerator: Leading-Color (n+1)-parton matrix element
            double antMELC = functionsPtr->A40(nPlusOneE,0,1,2,3);
            double antMENF = functionsPtr->B40(nPlusOneS,0,1,2,3);
            
            // Denominator: antenna approximation (start with this clustering)
            double antSumVinciaE   = ordE * aVinciaE * antOrg;
            double antSumVinciaS   = ordS * aVinciaS * antOrg;

            // Verbose output
            if (verbose >= 9) {
              cout<<"pNew\n";
              cout<<nPlusOneE[0]<<nPlusOneE[1]<<nPlusOneE[2]<<nPlusOneE[3];
              cout<<" antMELC = "<<scientific<<antMELC
                  <<" antSumVinciaE "<<iA+1<<" "<<antSumVinciaE<<endl;
              if (kineMapS != kineMapE)  
                cout<<nPlusOneS[0]<<nPlusOneS[1]<<nPlusOneS[2]<<nPlusOneS[3];
              cout<<" antMENF = "<<scientific<<antMENF
                  <<" antSumVinciaS "<<iA+1<<" "<<antSumVinciaS<<endl;
            }

            // Global showers: also sum over alternative clusterings
            for (int ir = 1; ir < int(nPlusOneE.size())-1; ++ir) {
              if (ir == iA+1 || sectorShower) continue;
              vector<Vec4> pClu(pOrg);                
              // === Gluon Emission ===
              // Cluster back using emission kine map
              vinCluPtr->map3to2(kineMapE, nPlusOneE[ir-1], nPlusOneE[ir], 
                                 nPlusOneE[ir+1], pClu[ir-1], pClu[ir]);
              // Compute ME for reclustered n-parton state
              // (here represented by 2->3 GGG antenna for Z->3)
              double antClu = functionsPtr->A30(pClu,0,1,2);
              // Compute 3->4 invariants for this clustering
              double sNow = m2(pClu[ir-1],pClu[ir]);
              vector<double> xNow;
              xNow.push_back(m2(nPlusOneE[ir-1],nPlusOneE[ir]) / sNow);
              xNow.push_back(m2(nPlusOneE[ir],nPlusOneE[ir+1]) / sNow);
              // Compute 3->4 antennae. Conjugate if needed.
              int iAntNow = iQGemit;
              bool swapNow = false;
              if (ir == 2) swapNow = true;
              Antenna* antPtrNow = antennaSets[0]->getAnt(iAntNow);
              double aVinciaNow  = 
                antPtrNow->antennaFunction(xNow[swapNow],xNow[1-swapNow])/sNow;
              // Compute ordering factor (for pT and mD, respectively)
              double mD2now = mDnorm * sNow * min(xNow[0],xNow[1]);
              double pT2now = sNow * pTnorm * xNow[0] * xNow[1];
              double q2now = (evTypeEmit == 1) ? pT2now : mD2now;
              double s01clu = 2*pClu[0]*pClu[1];
              double s12clu = 2*pClu[1]*pClu[2];
              double mD2clu = mDnorm * min(s01clu,s12clu);
              double pT2clu = pTnorm * s01clu * s12clu 
                / m2(pClu[0] + pClu[1] + pClu[2]);
              double q2clu  = (evTypeEmit == 1) ? pT2clu : mD2clu;
              double ordClu    = q2clu/(q2clu + q2now);
              antSumVinciaE   += ordClu * aVinciaNow * antClu;
              // === Gluon Splitting ===
              // Recompute invariants and clustered ME if necessary
              if (kineMapS != kineMapE) {
                // Cluster back using splitting kine map
                vinCluPtr->map3to2(kineMapS, nPlusOneS[ir-1], nPlusOneS[ir], 
                                   nPlusOneS[ir+1], pClu[ir-1], pClu[ir]);
                // Compute ME for reclustered n-parton state
                // (here represented by 2->3 GGG antenna for Z->3)
                antClu = functionsPtr->A30(pClu,0,1,2);
                // Compute 3->4 invariants for this clustering
                sNow = m2(pClu[ir-1],pClu[ir]);
                xNow.resize(0);
                xNow.push_back(m2(nPlusOneS[ir-1],nPlusOneS[ir]) / sNow);
                xNow.push_back(m2(nPlusOneS[ir],nPlusOneS[ir+1]) / sNow);
              }
              iAntNow = iQGsplit;
              swapNow = false;
              if (ir == 2) swapNow = true;
              // Compute antennae. Conjugate if needed.
              antPtrNow = antennaSets[0]->getAnt(iAntNow);
              aVinciaNow  = 
                antPtrNow->antennaFunction(xNow[swapNow],xNow[1-swapNow])/sNow;
              // Compute ordering factor (for pT and mD, respectively)
              q2now  = sNow * xNow[1-swapNow];
              mD2now = q2now;
              pT2now = sNow * pTnorm * x[0] * x[1];
              ordClu = q2clu/(q2clu + q2now);
              antSumVinciaS   += ordClu * aVinciaNow * antClu;
            }
            // Compute ME correction factors
            double RtreeE = antMELC/antSumVinciaE;
            double RtreeS = antMENF/antSumVinciaS;
            // Add to emit and split sums
            thisEmit  += ordE * (RtreeE*aVinciaE-aStdE);
            thisSplit += ordS * (RtreeS*aVinciaS-aStdS);
            
            if (verbose >= 9) 
              cout<<" Rtree Emit = "<<RtreeE<<"   "<<" Split = "<<RtreeS<<endl;
            
          } // end sum over antennae
          
          // Add this point to totals and compute approximation to deltaA
          sumEmit  += thisEmit; 
          sumSplit += thisSplit; 
          deltaAsq += pow2(comFacEmit*thisEmit+comFacSplit*thisSplit);
          deltaA   += comFacEmit*thisEmit + comFacSplit*thisSplit;

          // Evaluate precision. Require at least nPointsMCmin points
          if (nPointsMC < nPointsMCmin) 
            precision = 1.0;
          else if (nPointsMC >= nPointsMCmax) {
            double variance = 1./(nPointsMC-1) * (deltaAsq/nPointsMC
                                                  - pow2(deltaA/nPointsMC));
            // At end of loop, just check whether relative precision on 
            // deltaA is at least OK.
            precision = abs(variance/deltaA);
            // Otherwise print a warning message.
            if (precision > precisionGoal) 
              printOut("VM:getVME","Failed to converge on deltaA term");
            precision = 0.0;
          }
          else {
            // Normal operation: check absolute precision on deltaA correction
            // Multiply variance by 2, to check precision at 2-sigma level
            double variance = 1./(nPointsMC-1) * (deltaAsq/nPointsMC
                                                  - pow2(deltaA/nPointsMC));
            precision = abs(alphaSfac * 2 * variance);
          }
          
        } // end Monte Carlo integration loop
	
        // Restore VinClu verbose level
        vinCluPtr->setVerbose(verbose);

        // Add final deltaA term to shower Sudakov expansion
        shower       += deltaA / nPointsMC;

        // Matching correction (without renormalization term)
	double matchNLOnomuR = alphaSfac * ( shower + RfunFull );

	// Compute VME (including renormalization term)
	VME.resize(nVariations);
	VME[0] = matchNLOnomuR - SLmuR;        
        // Uncertainty variations
	if (uncertaintyBands) {
          VME[1] = matchNLOnomuR - SLmuRdef;
          VME[2] = matchNLOnomuR - SLmuRmin;
          VME[3] = matchNLOnomuR - SLmuRmax;
          VME[4] = VME[1];
          VME[5] = VME[1];
          VME[6] = VME[1];
          VME[7] = VME[1];
          // 8 and 9 : ordering variations: pT and mD, but Sudakov integrals 
          // are the same as for Def (regions not changed - yet)
          VME[8] = VME[1];
          VME[9] = VME[1];
          VME[10] = VME[1] * (1. + 1./pow2(NC));
          VME[11] = VME[1] / (1. + 1./pow2(NC));
	}
	status = true;

        // TMP: print terms and size of deltaA
        if ( verbose >= 8 || (verbose >= 3 && abs(VME[0]) > 0.5)) {
          cout<<" Large NLO correction for "<<particles[0].id()<<" -> ";
          for (int i=1; i<int(particles.size()); ++i) 
            cout<<particles[i].id()<<" ";
          cout<<endl;
          cout<<" NLO terms: shower = "<<shower<<" ( deltaA = "<<deltaA/nPointsMC<<" ariadneLog = "<<AriadneLog<<") ME = "<<RfunFull<<" ( = "<<RfunFull1<<" + "<<RfunFull2<<" + "<<RfunFull3<<" )"<<endl;
          cout<<" NLO corr = "<<num2str(VME[0],6)<<" ; size of deltaA = "<<num2str(deltaA/nPointsMC*alphaS/2./M_PI,6) << endl;

          printOut("VM::getVME","Shower expansion = "+num2str(shower));
          printOut("VM::getVME","   matchNLOnomuR = "+num2str(matchNLOnomuR));
          printOut("VM::getVME","Matching factor with muR=pT = "
                   +num2str(1.+VME[0]));
	}

        // Histogram relative numerical size of deltaA terms
        if ( verbose >= 3 ) {
          double rDeltaA = alphaSfac * deltaA/nPointsMC;
          int nQQtmp = 1;
          int nGtmp = 1;
          string state = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
          string hisTitle="DeltaA:" + state;
          HrDeltaA[hisTitle].fill(rDeltaA);
        }

      }
      
    }

    return status;

  }
  
  //*******

  // get2to4():
  // Return 2->4 reweighting factors 
  //   id  : id codes for the 4 partons
  //   p   : 4-momenta for the 4 partons
  // Return reweighting factors in PME
  //   P2to4[0] : reweighting factor with USR antennae
  //   P2to4[n] : reweighting factor with variations, see getPME()
  // = |A_4|^2 / (Sum_i A^phys_i A^phys_i_n-1)

  bool Matching::getP2to4(vector<int> id, vector<Vec4> p, 
			vector<double>& P2to4) {

    // Resize P2to4.
    P2to4.resize(0);
    bool done = false;

    // Check if at least 4 partons in final state
    if (id.size() <= 5) return false;

    // Initialize (n-1)->n vector of approximations
    vector<double> Mn2Appr(0);
    for (int iVar=0; iVar<nVariations; iVar++) {
      Mn2Appr.push_back(0.);
    }
    
    // SYSTEMATIC 2->4 MATCHING
    // Requires that collinear singularities correctly
    // partitioned in 2->4 functions

    // Loop over possible emitter antennae
    for (int iClu=0; iClu<int(id.size())-3; iClu++) {
      
      // Clustering:
      //cout<<" P2to4 Clustering "<<iClu<<"/"<<int(int(id.size())-4)<<endl;
      
      // Compute new ordering scale
      int ir  = 2+iClu;
      int ida = id[ir-1];
      int idr = id[ir];
      int idb = id[ir+1];
      
      //cout <<"  arb = "<<ir-1<<" "<<ir<<" "<<ir+1<<" QNew = "<<sqrt(Q2new)<<endl;
      
      // B40A: (q qbar) -> (q g) qbar -> q q'bar q' qbar
      // B40B: (q qbar) -> q (g qbar) -> q q'bar q' qbar
      
      // E40aA: (q g) -> (q qbar' q') -> q g qbar' q'
      // (match after gluon emission inside qqbar antenna)
      // E40aB: (q g) -> q (g g) -> q g qbar' q'
      // (match after gluon splitting inside gg antenna, outside)
      
      // E40bA: (q g) -> (q g) g -> q q'bar q' g
      // (match after gluon splitting inside qg antenna
      // E40bB: (q g) -> q (g g) -> q q'bar q' g
      // (match after gluon splitting inside gg antenna, inside)
      
      // H40A: (g g) -> qbar (q g) -> qbar q qbar' q'
      // (match after gluon splitting inside qg antenna
      // H40B: (g g) -> (g qbar) q -> qbar q qbar' q'
      // (match after gluon splitting inside gqbar antenna
      
      // G40A: (g g) -> (g g) g -> g qbar q g
      // (match after gluon splitting inside qg antenna
      // G40B: (g g) -> g (g g) -> g qbar q g
      // (match after gluon splitting inside gqbar antenna
      
      // Reconstruct previous state
      vector<int> idClu(id);
      vector<Vec4> pClu;
      int idaHat  = ida;
      int idbHat  = idb;
      int antIndx = -1;
      bool swapAB = false;
      if (idr == 21) {
	// To recluster gluon, just collapse parent ids
	idClu.erase(idClu.begin()+ir);
	if (ida == 21 && idb == 21) antIndx = iGGemit;
	else if (idb == 21) antIndx = iQGemit;
	else if (ida == 21) {antIndx = iQGemit; swapAB=true;}
	else antIndx = iQQemit;
      } else if (idr < 0) {
	// To recluster antiquark, collapse with downstream quark 
	idClu[ir+1]=21;
	idClu.erase(idClu.begin()+ir);
	idbHat = 21;
	if (idaHat == 21) antIndx = iGGsplit;
	else antIndx = iQGsplit;
      } else {
	// To recluster quark, collapse with upstream antiquark
	idClu[ir-1]=21;
	idClu.erase(idClu.begin()+ir);
	idaHat = 21;
	if (idbHat == 21) {antIndx = iGGsplit; swapAB=true;}
	else {antIndx = iQGsplit; swapAB=true;}
      }
      if (verbose >= 9) {
	cout << "P2to4 Clustering "<<iClu<<" : ";
	for (int i=0;i<int(idClu.size());i++) cout << " " << idClu[i];
	if (antIndx == iQGemit) cout << (swapAB ? " gqbarEmit" : " qgEmit");
	cout<<" Ant = "<<antennaSets[0]->getAnt(antIndx)->humanName();
	cout<<endl;
      }
      // Reconstruct clustered momenta
      int iMap = antennaSets[0]->getAnt(antIndx)->getKineMapType();
      vinCluPtr->vc3to2(iMap,p,ir-1,ir,ir+1,pClu);
      if (verbose >= 9) {
	for (int i=0;i<int(idClu.size());i++) {
	  cout << "   p" << i<<" "<<pClu[i].px()<<" "<<pClu[i].py()
	       <<" "<<pClu[i].pz()<<" "<<pClu[i].e()<<endl;
	}
      }
      
      // Invariant mass of parent antenna and daughter invariants
      double sP  = m2(pClu[ir-1],pClu[ir]);
      double yar = m2(p[ir-1],p[ir])/sP;
      double yrb = m2(p[ir],p[ir+1])/sP;
      if (swapAB) {
	double ytmp = yar;
	yar = yrb;
	yrb = ytmp;
      }
      
      // Compute Pimp factor (only used for global showers) 
      //  : orderingMode <= 1 : Pimp = 1
      //  : orderingMode == 2 : Q2E/(Q2Ehat+Q2E) 
      double Pimp    = 1.0;
      // Additional Pimp factor variations used for uncertainty estimates
      // (Not active in 2 to 4 yet)
      //double PimpStronger = 1.0;
      //double PimpMD  = 1.0;

      // Pimp factors only used in global showers
      if ( !sectorShower && orderingMode >= 2 ) { 

	// QE, pT, and mD Values for current clustering	
	double QE2new;
	// If this is an anti fermion -> always use m2rb for g->qqbar
	if (id[ir] < 0) {
	    QE2new = m2(p[ir],p[ir+1]);
	    //mD2new = QE2new;
	}
	// Else if this is a fermion -> always use m2ar for g->qqbar
	if (id[ir] < 0) {
	    QE2new = m2(p[ir-1],p[ir]);
	    //mD2new = QE2new;
	}
	// Else use QE, pT, mD definitions for g emission
	else {
	  QE2new = resolutionPtr->yE(yar,yrb)*sP;
	  //mD2new = resolutionPtr->yE(yar,yrb,2)*sP;
	}

	// Initial values for Hat variables characterizing the reclustered
	// parton state
	double QE2hat=m2(pClu[1],pClu[2],pClu[3]);
	//double mD2hat=mD2hat;
	// Loop over possible clusterings, find smallest Q for each measure
	for (int jClu=0; jClu<int(pClu.size())-3; jClu++) {
	  int jaHat=jClu+1;
	  int jrHat=jClu+2;
	  int jbHat=jClu+3;
	  // If this is an anti fermion -> use m2rb for g->qq splitting
	  if (idClu[jrHat] < 0 ) {
	    QE2hat = min(QE2hat,m2(pClu[jrHat],pClu[jbHat]));
	    //mD2hat = min(mD2hat,m2(pClu[jrHat],pClu[jbHat]));
	  }
	  // Else if this is a fermion, use m2ar for g->qq splitting
	  else if (idClu[jrHat] < 20) {
	    QE2hat = min(QE2hat,m2(pClu[jaHat],pClu[jrHat]));
	    //mD2hat = min(mD2hat,m2(pClu[jaHat],pClu[jrHat]));
	  }
	  // Else use individual QE, pT, or mD measure for g emission
	  else {
	    double QE2 = resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat]);
	    //double mD2 = resolutionPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],2);
	    QE2hat = min(QE2hat,QE2);
	    //mD2hat = min(mD2hat,mD2);
	  }
	}	  
	// Standard Pimp factor 
	Pimp  = QE2hat/(QE2hat+QE2new);
	// Pimp factor with quartic slope (used for uncertainty bands)
	//PimpStronger = pow2(QE2hat)/(pow2(QE2hat)+pow2(QE2new));
	// Pimp factor with MD2 for gluon emissions (used for uncertainty bands)
	//PimpMD  = mD2hat/(mD2hat+mD2new);
	if (verbose >= 8) 
	  printOut("VM::getP2to4","Pimp = "+num2str(Pimp));
      }
      
      // Compute ARIADNE factor (for g->qqbar splittings)
      double Pari = 1.0;
      if (!sectorShower && (antIndx == iQGsplit || antIndx == iGGsplit)) {
	double sN;
	if (!swapAB) {
	  // Only apply if neighbor is in final state
	  //	  if (ir+1 < int(pClu.size()) ) { 
	  sN = m2(pClu[ir],pClu[ir+1]);
	  // Only apply if neighbor is a quark
	  if (idClu[ir+1] != 21) 
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	} else { 
	  // Only apply if neighbor is in final state
	  //	  if (ir-2 >= nIn) {
	  sN = m2(pClu[ir-2],pClu[ir-1]);	
	  // Only apply if neighbor is a quark
	  if (idClu[ir-2] != 21)  
	    Pari = 2.0*sN/(sP + sN);
	  //	  }
	}
	
	// Verbose output
	if (verbose >= 8) 
	  printOut("Functions::getP2to4","Pari = "+num2str(Pari));
	
      }
      
      // Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
      double Mp2 = 0.0 ; // mgInterfacePtr->ME2(nIn, idClu, pClu);
      
      // TMP!
      if (verbose >= 9) {
	cout << " P2to4 Mp2 = "<<Mp2<<endl;
	
	// Test
	double sC01  = m2(pClu[1],pClu[2]);
	double sC12  = m2(pClu[2],pClu[3]);
	double sC    = m2(pClu[1],pClu[2],pClu[3]);
	double A3chk = CF*antennaSets[0]->getAnt(iQQemit)->
        antennaFunction(sC01/sC,sC12/sC)/sC;
	double A3chk2 = antennaSets[0]->getAnt(iQQemit)->
        getChargeFactor()
	  *antennaSets[0]->getAnt(iQQemit)->antennaFunction(sC01/sC,sC12/sC)/sC;
	cout << " P2to4 A30 = "<<scientific<<A3chk<<" s = "<<" A30chk = "
	     << A3chk2<<endl;
      }
      
      // Check if return was -1 ?
      if (abs(Mp2+1.0) < TINY) return false; 
      
      // Loop over different antenna sets (for uncertainty evaluation)
      // Use DEF set as basis for uncertainty variations
      vector<double> antVal;
      antVal.push_back(getAnt(antIndx,sectorShower,0)->
                        antennaFunction(yar,yrb)/sP); 
      if (uncertaintyBands) {
        antVal.push_back(getAnt(antIndx,sectorShower,1)->
                          antennaFunction(yar,yrb)/sP); 
        antVal.push_back(getAnt(antIndx,sectorShower,2)->
                          antennaFunction(yar,yrb)/sP); 
        antVal.push_back(getAnt(antIndx,sectorShower,3)->
                          antennaFunction(yar,yrb)/sP); 
      }
      for (int iVar=0; iVar<nVariations; iVar++) {	
	// Compute physical antenna for this set for this (n-1)-parton state
	double antPhys = antVal[0];
        if (iVar >= 1) antPhys = antVal[1];
        if (iVar == 4) antPhys = antVal[2];
        if (iVar == 5) antPhys = antVal[3];
	double chargeFactor = getAnt(antIndx,sectorShower,0)->getChargeFactor();
        if (iVar >= 1) 
          chargeFactor = getAnt(antIndx,sectorShower,1)->getChargeFactor();
	if (iVar == 10 && id[ir] == 21) chargeFactor = CA;
	if (iVar == 11 && id[ir] == 21) chargeFactor = CF;
	antPhys *= chargeFactor;
	antPhys *= Pimp * Pari;
	// Add to Approximated Mn2 for this set
	Mn2Appr[iVar] += Mp2 * antPhys;	
      }
      
    }
    
    // Store ME reweight factors
    for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {
      if (iSet >= 1 && ! uncertaintyBands) continue;
      //P2to4.push_back(Mn2/Mn2Appr[iSet]);
    }
    
    // Store Old ME
    //oldME2 = Mn2;
    
    
    // Done
    done = true;
    
    // For the time being, no non-MadGraph matching implemented
    return done;

  }

}
